"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const XRegExp = require("xregexp");
const gensequence_1 = require("gensequence");
const Text = require("./text");
const fileReader_1 = require("./fileReader");
const cspell_lib_1 = require("cspell-lib");
const rxjs_1 = require("rxjs");
const operators_1 = require("rxjs/operators");
const path = require("path");
const fs_extra_1 = require("fs-extra");
const Trie = require("cspell-trie");
const HR = require("hunspell-reader");
const regNonWordOrSpace = XRegExp("[^\\p{L}' ]+", 'gi');
const regExpSpaceOrDash = /(?:\s+)|(?:-+)/g;
const regExpRepeatChars = /(.)\1{3,}/i;
function normalizeWords(lines) {
    return lines.pipe(operators_1.flatMap(line => lineToWords(line)));
}
exports.normalizeWords = normalizeWords;
function lineToWords(line) {
    // Remove punctuation and non-letters.
    const filteredLine = line.replace(regNonWordOrSpace, '|');
    const wordGroups = filteredLine.split('|');
    const words = gensequence_1.genSequence(wordGroups)
        .concatMap(a => [a, ...a.split(regExpSpaceOrDash)])
        .concatMap(a => splitCamelCase(a))
        .map(a => a.trim())
        .filter(a => !!a)
        .filter(s => !regExpRepeatChars.test(s))
        .map(a => a.toLowerCase());
    return words;
}
exports.lineToWords = lineToWords;
function splitCamelCase(word) {
    const splitWords = Text.splitCamelCaseWord(word);
    // We only want to preserve this: "New York" and not "Namespace DNSLookup"
    if (splitWords.length > 1 && regExpSpaceOrDash.test(word)) {
        return gensequence_1.genSequence(splitWords).concatMap(w => w.split(regExpSpaceOrDash));
    }
    return splitWords;
}
function compileWordList(filename, destFilename, options) {
    return options.splitWords
        ? compileWordListWithSplit(filename, destFilename)
        : compileSimpleWordList(filename, destFilename);
}
exports.compileWordList = compileWordList;
function compileWordListWithSplit(filename, destFilename) {
    const getWords = () => regHunspellFile.test(filename) ? readHunspellFiles(filename) : fileReader_1.lineReaderRx(filename);
    const destDir = path.dirname(destFilename);
    return fs_extra_1.mkdirp(destDir).then(() => cspell_lib_1.writeToFileRxP(destFilename, getWords().pipe(operators_1.flatMap(line => lineToWords(line).toArray()), operators_1.distinct(), operators_1.map(a => a + '\n'), operators_1.bufferCount(1024), operators_1.map(a => a.join('')))));
}
exports.compileWordListWithSplit = compileWordListWithSplit;
function compileSimpleWordList(filename, destFilename) {
    const getWords = () => regHunspellFile.test(filename) ? readHunspellFiles(filename) : fileReader_1.lineReaderRx(filename);
    const destDir = path.dirname(destFilename);
    return fs_extra_1.mkdirp(destDir).then(() => cspell_lib_1.writeToFileRxP(destFilename, getWords().pipe(operators_1.map(a => a.toLowerCase()), operators_1.distinct(), operators_1.filter(a => !!a), operators_1.map(a => a + '\n'), operators_1.bufferCount(1024), operators_1.map(a => a.join('')))));
}
exports.compileSimpleWordList = compileSimpleWordList;
function normalizeWordsToTrie(words) {
    const result = normalizeWords(words)
        .pipe(operators_1.reduce((node, word) => Trie.insert(word, node), {}))
        .toPromise();
    return result;
}
exports.normalizeWordsToTrie = normalizeWordsToTrie;
function compileWordListToTrieFile(words, destFilename) {
    const destDir = path.dirname(destFilename);
    const dir = fs_extra_1.mkdirp(destDir);
    const root = normalizeWordsToTrie(words);
    const data = rxjs_1.zip(dir, root, (_, b) => b).pipe(operators_1.map(node => Trie.serializeTrie(node, { base: 32, comment: 'Built by cspell-tools.' })), operators_1.flatMap(seq => rxjs_1.from(seq)));
    return cspell_lib_1.writeToFileRxP(destFilename, data.pipe(operators_1.bufferCount(1024), operators_1.map(a => a.join(''))));
}
exports.compileWordListToTrieFile = compileWordListToTrieFile;
const regHunspellFile = /\.(dic|aff)$/i;
function readHunspellFiles(filename) {
    const dicFile = filename.replace(regHunspellFile, '.dic');
    const affFile = filename.replace(regHunspellFile, '.aff');
    const reader = HR.HunspellReader.createFromFiles(affFile, dicFile);
    const r = rxjs_1.from(reader).pipe(operators_1.flatMap(reader => reader.readWordsRx()), operators_1.map(aff => aff.word));
    return r;
}
function compileTrie(filename, destFilename) {
    const words = regHunspellFile.test(filename) ? readHunspellFiles(filename) : fileReader_1.lineReaderRx(filename);
    return compileWordListToTrieFile(words, destFilename);
}
exports.compileTrie = compileTrie;
//# sourceMappingURL=wordListCompiler.js.map